@@ -1,4 +1,4 @@
-// Copyright (C) 2018-2021 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2018-2023 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -56,16 +56,14 @@ enum DbMessageID {
     PGSQL_START_TRANSACTION,
     PGSQL_COMMIT,
     PGSQL_ROLLBACK,
+    PGSQL_CREATE_SAVEPOINT,
+    PGSQL_ROLLBACK_SAVEPOINT,
+    PGSQL_TCP_USER_TIMEOUT_UNSUPPORTED,
 
     MYSQL_FATAL_ERROR,
     MYSQL_START_TRANSACTION,
     MYSQL_COMMIT,
     MYSQL_ROLLBACK,
-
-    CQL_DEALLOC_ERROR,
-    CQL_CONNECTION_BEGIN_TRANSACTION,
-    CQL_CONNECTION_COMMIT,
-    CQL_CONNECTION_ROLLBACK
 };
 ///@}
 
