@@ -1,4 +1,4 @@
-// Copyright (C) 2017-2019 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2017-2023 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -121,7 +121,7 @@ public:
 
     /// @brief Returns a pointer to the collection of subnets within this
     /// shared network.
-    const Subnet4Collection* getAllSubnets() const {
+    const Subnet4SimpleCollection* getAllSubnets() const {
         return (&subnets_);
     }
 
@@ -195,24 +195,27 @@ public:
     bool subnetsIncludeMatchClientId(const Subnet4Ptr& first_subnet,
                                      const ClientClasses& client_classes);
 
-    /// @brief Check if the shared network includes a subnet with
-    /// not global host reservation mode.
-    ///
-    /// @return First subnet which has not a global host reservation mode.
-    Subnet4Ptr subnetsAllHRGlobal() const;
-
     /// @brief Unparses shared network object.
     ///
     /// @return A pointer to unparsed shared network configuration.
     virtual data::ElementPtr toElement() const;
 
+    /// @brief Generates an identifying label for logging.
+    ///
+    /// @return string containing the label
+    virtual std::string getLabel() const {
+        std::stringstream ss;
+        ss << "shared-network " << name_;
+        return (ss.str());
+    }
+
 private:
 
     /// @brief Holds a name of a shared network.
     std::string name_;
 
     /// @brief Collection of IPv4 subnets within shared network.
-    Subnet4Collection subnets_;
+    Subnet4SimpleCollection subnets_;
 };
 
 /// @brief Multi index container holding shared networks.
@@ -337,7 +340,7 @@ public:
 
     /// @brief Returns a pointer to the collection of subnets within this
     /// shared network.
-    const Subnet6Collection* getAllSubnets() const {
+    const Subnet6SimpleCollection* getAllSubnets() const {
         return (&subnets_);
     }
 
@@ -399,24 +402,27 @@ public:
     Subnet6Ptr getPreferredSubnet(const Subnet6Ptr& selected_subnet,
                                   const Lease::Type& lease_type) const;
 
-    /// @brief Check if the shared network includes a subnet with
-    /// not global host reservation mode.
-    ///
-    /// @return First subnet which has not a global host reservation mode.
-    Subnet6Ptr subnetsAllHRGlobal() const;
-
     /// @brief Unparses shared network object.
     ///
     /// @return A pointer to unparsed shared network configuration.
     virtual data::ElementPtr toElement() const;
 
+    /// @brief Generates an identifying label for logging.
+    ///
+    /// @return string containing the label
+    virtual std::string getLabel() const {
+        std::stringstream ss;
+        ss << "shared-network " << name_;
+        return (ss.str());
+    }
+
 private:
 
     /// @brief Holds a name of a shared network.
     std::string name_;
 
     /// @brief Collection of IPv6 subnets within shared network.
-    Subnet6Collection subnets_;
+    Subnet6SimpleCollection subnets_;
 };
 
 /// @brief Multi index container holding shared networks.
