@@ -1,4 +1,4 @@
-// Copyright (C) 2012-2021 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2012-2023 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -11,11 +11,17 @@
 #include <cc/data.h>
 #include <cc/user_context.h>
 #include <dhcp/option_space_container.h>
+#include <dhcpsrv/allocator.h>
+#include <dhcpsrv/allocation_state.h>
 #include <dhcpsrv/lease.h>
 #include <dhcpsrv/network.h>
 #include <dhcpsrv/pool.h>
 #include <dhcpsrv/subnet_id.h>
-#include <dhcpsrv/triplet.h>
+#include <util/bigints.h>
+#include <util/dhcp_space.h>
+#include <util/triplet.h>
+
+#include <boost/enable_shared_from_this.hpp>
 #include <boost/multi_index/mem_fun.hpp>
 #include <boost/multi_index/indexed_by.hpp>
 #include <boost/multi_index/ordered_index.hpp>
@@ -23,11 +29,10 @@
 #include <boost/multi_index_container.hpp>
 #include <boost/date_time/posix_time/posix_time.hpp>
 #include <boost/pointer_cast.hpp>
-#include <boost/scoped_ptr.hpp>
 #include <boost/shared_ptr.hpp>
+
 #include <cstdint>
 #include <map>
-#include <mutex>
 #include <utility>
 
 namespace isc {
@@ -72,54 +77,12 @@ public:
                 const isc::asiolink::IOAddress& addr,
                 const ClientClasses& client_classes) const;
 
-    /// @brief returns the last address that was tried from this subnet.
-    ///
-    /// This method returns the last address that was attempted to be allocated
-    /// from this subnet. This is used as helper information for the next
-    /// iteration of the allocation algorithm.
-    ///
-    /// @note: this routine is Kea thread safe.
-    ///
-    /// @todo: Define map<SubnetID, ClientClass, IOAddress> somewhere in the
-    ///        AllocEngine::IterativeAllocator and keep the data there
-    ///
-    /// @param type lease type to be returned
-    /// @return address/prefix that was last tried from this subnet
-    isc::asiolink::IOAddress getLastAllocated(Lease::Type type) const;
-
-    /// @brief Returns the timestamp when the @c setLastAllocated function
-    /// was called.
-    ///
-    /// @note: this routine is Kea thread safe.
-    ///
-    /// @param lease_type Lease type for which last allocation timestamp should
-    /// be returned.
-    ///
-    /// @return Time when a lease of a specified type has been allocated from
-    /// this subnet. The negative infinity time is returned if a lease type is
-    /// not recognized (which is unlikely).
-    boost::posix_time::ptime
-    getLastAllocatedTime(const Lease::Type& lease_type) const;
-
-    /// @brief sets the last address that was tried from this subnet.
-    ///
-    /// This method sets the last address that was attempted to be allocated
-    /// from this subnet. This is used as helper information for the next
-    /// iteration of the allocation algorithm.
-    ///
-    /// @note: this routine is Kea thread safe.
-    ///
-    /// @todo: Define map<SubnetID, ClientClass, IOAddress> somewhere in the
-    ///        AllocEngine::IterativeAllocator and keep the data there
-    /// @param addr address/prefix to that was tried last
-    /// @param type lease type to be set
-    void setLastAllocated(Lease::Type type,
-                          const isc::asiolink::IOAddress& addr);
-
-    /// @brief Returns unique ID for that subnet.
+     /// @brief Returns unique ID for that subnet.
     ///
     /// @return unique ID for that subnet
-    SubnetID getID() const { return (id_); }
+    SubnetID getID() const {
+        return (id_);
+    }
 
     /// @brief Returns subnet parameters (prefix and prefix length).
     ///
@@ -146,10 +109,9 @@ public:
     ///
     /// @param pool pool to be added
     ///
-    /// @throw isc::BadValue if the pool type is invalid, the pool
-    /// is not an IA_PD pool and the address range of this pool does not
-    /// match the subnet prefix, or the pool overlaps with an existing pool
-    /// within the subnet.
+    /// @throw isc::BadValue if the pool type is invalid, in the case of address
+    /// pools if the address range of the pool does not match the subnet prefix,
+    /// or if the pool overlaps with an existing pool within the subnet.
     void addPool(const PoolPtr& pool);
 
     /// @brief Deletes all pools of specified type.
@@ -168,7 +130,7 @@ public:
     /// If there is no pool that the address belongs to (hint is invalid), other
     /// pool of specified type will be returned.
     ///
-    /// With anypool set to true, this is means give me a pool, preferably
+    /// With anypool set to true, this means give me a pool, preferably
     /// the one that addr belongs to. With anypool set to false, it means
     /// give me a pool that addr belongs to (or NULL if here is no such pool)
     ///
@@ -216,7 +178,7 @@ public:
     /// @brief Returns the number of possible leases for specified lease type.
     ///
     /// @param type type of the lease
-    uint64_t getPoolCapacity(Lease::Type type) const;
+    isc::util::uint128_t getPoolCapacity(Lease::Type type) const;
 
     /// @brief Returns the number of possible leases for specified lease type
     /// allowed for a client which belongs to classes.
@@ -224,8 +186,24 @@ public:
     /// @param type type of the lease
     /// @param client_classes list of classes the client belongs to
     /// @return number of leases matching lease type and classes
-    uint64_t getPoolCapacity(Lease::Type type,
-                             const ClientClasses& client_classes) const;
+    isc::util::uint128_t getPoolCapacity(Lease::Type type,
+                                         const ClientClasses& client_classes) const;
+
+    /// @brief Returns the number of possible leases for specified lease type
+    /// allowed for a client which belongs to classes and matching selection
+    /// criteria relative to provided hint prefix length.
+    ///
+    /// @param type type of the lease
+    /// @param client_classes list of classes the client belongs to
+    /// @param prefix_length_match type which indicates the selection criteria
+    /// for the pools relative to the provided hint prefix length
+    /// @param hint_prefix_length the hint prefix length that the client
+    /// provided
+    /// @return number of leases matching lease type and classes
+    isc::util::uint128_t getPoolCapacity(Lease::Type type,
+                                         const ClientClasses& client_classes,
+                                         Allocator::PrefixLenMatchType prefix_length_match,
+                                         uint8_t hint_prefix_length) const;
 
     /// @brief Returns textual representation of the subnet (e.g.
     /// "2001:db8::/64").
@@ -306,6 +284,53 @@ public:
     /// @return a collection of all pools
     PoolCollection& getPoolsWritable(Lease::Type type);
 
+    /// @brief Returns lease allocator instance.
+    ///
+    /// An allocator is responsible for selecting leases from the subnet's
+    /// pools. Each subnet has one allocator common for all pools belonging
+    /// to the subnet. The allocation engine uses this function to get the
+    /// current subnet allocator and uses it to select and offer an address.
+    ///
+    /// @param type lease type for which the allocator instance should be
+    /// returned.
+    /// @return Allocator instance.
+    AllocatorPtr getAllocator(Lease::Type type) const;
+
+    /// @brief Sets new allocator instance.
+    ///
+    /// If the server is configured to use a different allocator for the
+    /// subnet, it can set the current allocator with this function.
+    ///
+    /// @param type lease type for which the allocator is set.
+    /// @param allocator new allocator instance.
+    void setAllocator(Lease::Type type, const AllocatorPtr& allocator);
+
+    /// @brief Returns subnet-specific allocation state.
+    ///
+    /// The actual type of the state depends on the allocator type.
+    ///
+    /// @param type lease type for which the allocation state is returned.
+    /// @return allocation state.
+    SubnetAllocationStatePtr getAllocationState(Lease::Type type) const;
+
+    /// @brief Sets subnet-specific allocation state.
+    ///
+    /// @param type lease type for which the allocation state is set.
+    /// @param allocation_state allocation state instance.
+    void setAllocationState(Lease::Type type, const SubnetAllocationStatePtr& allocation_state);
+
+    /// @brief Instantiates the allocators and their states.
+    ///
+    /// It determines the types of the allocators to create using the list of
+    /// the allocator types specified with the @c Network::setAllocatorType method.
+    ///
+    /// This function is called from the subnet parsers and after fetching
+    /// the subnet configuration from a configuration backend.
+    virtual void createAllocators() = 0;
+
+    /// @brief Calls @c initAfterConfigure for each allocator.
+    void initAllocatorsAfterConfigure();
+
 protected:
 
     /// @brief Protected constructor.
@@ -370,15 +395,32 @@ protected:
     ///
     /// @param pools list of pools
     /// @return sum of possible leases
-    uint64_t sumPoolCapacity(const PoolCollection& pools) const;
+    isc::util::uint128_t sumPoolCapacity(const PoolCollection& pools) const;
 
     /// @brief Returns a sum of possible leases in all pools allowing classes.
     ///
     /// @param pools list of pools
     /// @param client_classes list of classes
     /// @return sum of possible/allowed leases
-    uint64_t sumPoolCapacity(const PoolCollection& pools,
-                             const ClientClasses& client_classes) const;
+    isc::util::uint128_t sumPoolCapacity(const PoolCollection& pools,
+                                         const ClientClasses& client_classes) const;
+
+    /// @brief Returns a sum of possible leases in all pools allowing classes
+    /// and matching selection criteria relative to provided hint prefix length.
+    ///
+    /// @note This function should be called only for PD pools.
+    ///
+    /// @param pools list of pools
+    /// @param client_classes list of classes
+    /// @param prefix_length_match type which indicates the selection criteria
+    /// for the pools relative to the provided hint prefix length
+    /// @param hint_prefix_length the hint prefix length that the client
+    /// provided
+    /// @return sum of possible/allowed leases
+    isc::util::uint128_t sumPoolCapacity(const PoolCollection& pools,
+                                         const ClientClasses& client_classes,
+                                         Allocator::PrefixLenMatchType prefix_length_match,
+                                         uint8_t hint_prefix_length) const;
 
     /// @brief Checks if the specified pool overlaps with an existing pool.
     ///
@@ -395,6 +437,12 @@ protected:
     /// @return A pointer to unparsed subnet configuration.
     virtual data::ElementPtr toElement() const;
 
+    virtual std::string getLabel() const {
+        std::stringstream ss;
+        ss << "subnet-id " << id_;
+        return (ss.str());
+    }
+
     /// @brief Converts subnet prefix to a pair of prefix/length pair.
     ///
     /// IPv4 and IPv6 specific conversion functions should apply extra checks
@@ -427,86 +475,14 @@ protected:
     /// @brief a prefix length of the subnet.
     uint8_t prefix_len_;
 
-    /// @brief last allocated address.
-    ///
-    /// This is the last allocated address that was previously allocated from
-    /// this particular subnet. Some allocation algorithms (e.g. iterative) use
-    /// that value, others do not. It should be noted that although the value
-    /// is usually correct, there are cases when it is invalid, e.g. after
-    /// removing a pool, restarting or changing allocation algorithms. For
-    /// that purpose it should be only considered a help that should not be
-    /// fully trusted.
-    isc::asiolink::IOAddress last_allocated_ia_;
-
-    /// @brief last allocated temporary address.
-    ///
-    /// See @ref last_allocated_ia_ for details.
-    isc::asiolink::IOAddress last_allocated_ta_;
-
-    /// @brief last allocated IPv6 prefix.
-    ///
-    /// See @ref last_allocated_ia_ for details.
-    isc::asiolink::IOAddress last_allocated_pd_;
-
-    /// @brief Timestamp indicating when a lease of a specified type has been
-    /// last allocated from this subnet.
-    ///
-    /// @note: This map is protected by the mutex.
-    std::map<Lease::Type, boost::posix_time::ptime> last_allocated_time_;
-
     /// @brief Shared network name.
     std::string shared_network_name_;
 
-private:
+    /// @brief Lease allocators used by the subnet.
+    std::map<Lease::Type, AllocatorPtr> allocators_;
 
-    /// @brief returns the last address that was tried from this subnet.
-    ///
-    /// Should be called in a thread safe context.
-    ///
-    /// This method returns the last address that was attempted to be allocated
-    /// from this subnet. This is used as helper information for the next
-    /// iteration of the allocation algorithm.
-    ///
-    /// @todo: Define map<SubnetID, ClientClass, IOAddress> somewhere in the
-    ///        AllocEngine::IterativeAllocator and keep the data there
-    ///
-    /// @param type lease type to be returned
-    /// @return address/prefix that was last tried from this subnet
-    isc::asiolink::IOAddress getLastAllocatedInternal(Lease::Type type) const;
-
-    /// @brief Returns the timestamp when the @c setLastAllocated function
-    /// was called.
-    ///
-    /// Should be called in a thread safe context.
-    ///
-    /// @param lease_type Lease type for which last allocation timestamp should
-    /// be returned.
-    ///
-    /// @return Time when a lease of a specified type has been allocated from
-    /// this subnet. The negative infinity time is returned if a lease type is
-    /// not recognized (which is unlikely).
-    boost::posix_time::ptime
-    getLastAllocatedTimeInternal(const Lease::Type& lease_type) const;
-
-    /// @brief sets the last address that was tried from this subnet.
-    ///
-    /// Should be called in a thread safe context.
-    ///
-    /// This method sets the last address that was attempted to be allocated
-    /// from this subnet. This is used as helper information for the next
-    /// iteration of the allocation algorithm.
-    ///
-    /// @note: this routine is Kea thread safe.
-    ///
-    /// @todo: Define map<SubnetID, ClientClass, IOAddress> somewhere in the
-    ///        AllocEngine::IterativeAllocator and keep the data there
-    /// @param addr address/prefix to that was tried last
-    /// @param type lease type to be set
-    void setLastAllocatedInternal(Lease::Type type,
-                                  const isc::asiolink::IOAddress& addr);
-
-    /// @brief Mutex to protect the internal state.
-    boost::scoped_ptr<std::mutex> mutex_;
+    /// @brief Holds subnet-specific allocation state.
+    std::map<Lease::Type, SubnetAllocationStatePtr> allocation_states_;
 };
 
 /// @brief A generic pointer to either Subnet4 or Subnet6 object
@@ -526,7 +502,7 @@ typedef boost::shared_ptr<Subnet4> Subnet4Ptr;
 /// This class represents an IPv4 subnet.
 /// @note Subnet and Network use virtual inheritance to avoid
 /// a diamond issue with UserContext
-class Subnet4 : public Subnet, public Network4 {
+class Subnet4 : public Subnet, public Network4, public boost::enable_shared_from_this<Subnet4> {
 public:
 
     /// @brief Constructor with all parameters.
@@ -538,13 +514,12 @@ public:
     /// @param t1 renewal timer (in seconds)
     /// @param t2 rebind timer (in seconds)
     /// @param valid_lifetime preferred lifetime of leases (in seconds)
-    /// @param id arbitrary subnet id, default value of 0 triggers
-    /// autogeneration of subnet id
+    /// @param id arbitrary subnet id, no default value
     Subnet4(const isc::asiolink::IOAddress& prefix, uint8_t length,
-            const Triplet<uint32_t>& t1,
-            const Triplet<uint32_t>& t2,
-            const Triplet<uint32_t>& valid_lifetime,
-            const SubnetID id = 0);
+            const util::Triplet<uint32_t>& t1,
+            const util::Triplet<uint32_t>& t2,
+            const util::Triplet<uint32_t>& valid_lifetime,
+            const SubnetID id);
 
     /// @brief Factory function creating an instance of the @c Subnet4.
     ///
@@ -553,21 +528,24 @@ public:
     /// unloaded before the object is destroyed. This ensures that the
     /// ownership of the object by the Kea process is retained.
     ///
+    /// It associates the subnet with the default, iterative, allocator.
+    /// Therefore, using this function should be preferred over the
+    /// constructor whenever the subnet needs a default allocator.
+    ///
     /// @param prefix Subnet4 prefix
     /// @param length prefix length
     /// @param t1 renewal timer (in seconds)
     /// @param t2 rebind timer (in seconds)
     /// @param valid_lifetime preferred lifetime of leases (in seconds)
-    /// @param id arbitrary subnet id, default value of 0 triggers
-    /// autogeneration of subnet id
+    /// @param id arbitrary subnet id, no default value
     ///
     /// @return Pointer to the @c Subnet4 instance.
     static Subnet4Ptr
     create(const isc::asiolink::IOAddress& prefix, uint8_t length,
-           const Triplet<uint32_t>& t1,
-           const Triplet<uint32_t>& t2,
-           const Triplet<uint32_t>& valid_lifetime,
-           const SubnetID id = 0);
+           const util::Triplet<uint32_t>& t1,
+           const util::Triplet<uint32_t>& t2,
+           const util::Triplet<uint32_t>& valid_lifetime,
+           const SubnetID id);
 
     /// @brief Returns next subnet within shared network.
     ///
@@ -633,6 +611,15 @@ public:
     /// @return A pointer to unparsed subnet configuration.
     virtual data::ElementPtr toElement() const;
 
+    /// @brief Instantiates the allocator and its state.
+    ///
+    /// It uses the type of the allocator specified with the
+    /// @c Network::setAllocatorType method.
+    ///
+    /// This function is called from the subnet parsers and after fetching
+    /// the subnet configuration from a configuration backend.
+    virtual void createAllocators();
+
     /// @brief Converts subnet prefix to a pair of prefix/length pair.
     ///
     /// @param prefix Prefix to be parsed.
@@ -642,6 +629,12 @@ public:
 
 private:
 
+    /// @brief Deleted copy constructor.
+    Subnet4(const Subnet4&) = delete;
+
+    /// @brief Deleted assignment operator.
+    Subnet4& operator=(const Subnet4&) = delete;
+
     /// @brief Returns default address for pool selection.
     ///
     /// @return ANY IPv4 address
@@ -674,7 +667,7 @@ typedef boost::shared_ptr<Subnet6> Subnet6Ptr;
 /// This class represents an IPv6 subnet.
 /// @note Subnet and Network use virtual inheritance to avoid
 /// a diamond issue with UserContext
-class Subnet6 : public Subnet, public Network6 {
+class Subnet6 : public Subnet, public Network6, public boost::enable_shared_from_this<Subnet6> {
 public:
 
     /// @brief Constructor with all parameters.
@@ -687,14 +680,14 @@ public:
     /// @param t2 rebind timer (in seconds)
     /// @param preferred_lifetime preferred lifetime of leases (in seconds)
     /// @param valid_lifetime preferred lifetime of leases (in seconds)
-    /// @param id arbitrary subnet id, default value of 0 triggers
+    /// @param id arbitrary subnet id, no default value
     /// autogeneration of subnet id
     Subnet6(const isc::asiolink::IOAddress& prefix, uint8_t length,
-            const Triplet<uint32_t>& t1,
-            const Triplet<uint32_t>& t2,
-            const Triplet<uint32_t>& preferred_lifetime,
-            const Triplet<uint32_t>& valid_lifetime,
-            const SubnetID id = 0);
+            const util::Triplet<uint32_t>& t1,
+            const util::Triplet<uint32_t>& t2,
+            const util::Triplet<uint32_t>& preferred_lifetime,
+            const util::Triplet<uint32_t>& valid_lifetime,
+            const SubnetID id);
 
     /// @brief Factory function creating an instance of the @c Subnet4.
     ///
@@ -703,23 +696,26 @@ public:
     /// unloaded before the object is destroyed. This ensures that the
     /// ownership of the object by the Kea process is retained.
     ///
+    /// It associates the subnet with the default, iterative, allocator.
+    /// Therefore, using this function should be preferred over the
+    /// constructor whenever the subnet needs a default allocator.
+    ///
     /// @param prefix Subnet6 prefix
     /// @param length prefix length
     /// @param t1 renewal timer (in seconds)
     /// @param t2 rebind timer (in seconds)
     /// @param preferred_lifetime preferred lifetime of leases (in seconds)
     /// @param valid_lifetime preferred lifetime of leases (in seconds)
-    /// @param id arbitrary subnet id, default value of 0 triggers
-    /// autogeneration of subnet id
+    /// @param id arbitrary subnet id, no default value
     ///
     /// @return Pointer to the @c Subnet6 instance.
     static Subnet6Ptr
     create(const isc::asiolink::IOAddress& prefix, uint8_t length,
-           const Triplet<uint32_t>& t1,
-           const Triplet<uint32_t>& t2,
-           const Triplet<uint32_t>& preferred_lifetime,
-           const Triplet<uint32_t>& valid_lifetime,
-           const SubnetID id = 0);
+           const util::Triplet<uint32_t>& t1,
+           const util::Triplet<uint32_t>& t2,
+           const util::Triplet<uint32_t>& preferred_lifetime,
+           const util::Triplet<uint32_t>& valid_lifetime,
+           const SubnetID id);
 
     /// @brief Returns next subnet within shared network.
     ///
@@ -764,6 +760,15 @@ public:
     virtual bool
     clientSupported(const isc::dhcp::ClientClasses& client_classes) const;
 
+    /// @brief Instantiates the allocators and their states.
+    ///
+    /// It determines the types of the allocators to create using the list of
+    /// the allocator types specified with the @c Network::setAllocatorType method.
+    ///
+    /// This function is called from the subnet parsers and after fetching
+    /// the subnet configuration from a configuration backend.
+    virtual void createAllocators();
+
     /// @brief Unparse a subnet object.
     ///
     /// @return A pointer to unparsed subnet configuration.
@@ -778,7 +783,14 @@ public:
 
 private:
 
+    /// @brief Deleted copy constructor.
+    Subnet6(const Subnet6&) = delete;
+
+    /// @brief Deleted assignment operator.
+    Subnet6& operator=(const Subnet6&) = delete;
+
     /// @brief Returns default address for pool selection
+    ///
     /// @return ANY IPv6 address
     virtual isc::asiolink::IOAddress default_pool() const {
         return (isc::asiolink::IOAddress("::"));
@@ -810,6 +822,30 @@ struct SubnetServerIdIndexTag { };
 /// @brief Tag for the index for searching by subnet modification time.
 struct SubnetModificationTimeIndexTag { };
 
+/// @brief A simple collection of @c Subnet4 objects
+///
+/// This container provides a set of indexes which can be used to retrieve
+/// subnets by subnet identifier and subnet prefix.
+///
+/// The random access index is used for direct iteration over the collection.
+typedef boost::multi_index_container<
+    // Multi index container holds pointers to the subnets.
+    Subnet4Ptr,
+    // The following holds all indexes.
+    boost::multi_index::indexed_by<
+        // First index allows for searching using subnet identifier.
+        boost::multi_index::ordered_unique<
+            boost::multi_index::tag<SubnetSubnetIdIndexTag>,
+            boost::multi_index::const_mem_fun<Subnet, SubnetID, &Subnet::getID>
+        >,
+        // Second index allows for searching using an output from toText function.
+        boost::multi_index::ordered_unique<
+            boost::multi_index::tag<SubnetPrefixIndexTag>,
+            boost::multi_index::const_mem_fun<Subnet, std::string, &Subnet::toText>
+        >
+    >
+> Subnet4SimpleCollection;
+
 /// @brief A collection of @c Subnet4 objects.
 ///
 /// This container provides a set of indexes which can be used to retrieve
@@ -866,6 +902,30 @@ typedef boost::multi_index_container<
     >
 > Subnet4Collection;
 
+/// @brief A simple collection of @c Subnet6 objects
+///
+/// This container provides a set of indexes which can be used to retrieve
+/// subnets by subnet identifier and subnet prefix.
+///
+/// The random access index is used for direct iteration over the collection.
+typedef boost::multi_index_container<
+    // Multi index container holds pointers to the subnets.
+    Subnet6Ptr,
+    // The following holds all indexes.
+    boost::multi_index::indexed_by<
+        // First index allows for searching using subnet identifier.
+        boost::multi_index::ordered_unique<
+            boost::multi_index::tag<SubnetSubnetIdIndexTag>,
+            boost::multi_index::const_mem_fun<Subnet, SubnetID, &Subnet::getID>
+        >,
+        // Second index allows for searching using an output from toText function.
+        boost::multi_index::ordered_unique<
+            boost::multi_index::tag<SubnetPrefixIndexTag>,
+            boost::multi_index::const_mem_fun<Subnet, std::string, &Subnet::toText>
+        >
+    >
+> Subnet6SimpleCollection;
+
 /// @brief A collection of @c Subnet6 objects
 ///
 /// This container provides a set of indexes which can be used to retrieve
@@ -918,7 +978,8 @@ typedef boost::multi_index_container<
 ///
 /// @tparam ReturnPtrType Type of the returned object, i.e. @c Subnet4Ptr
 /// or @c Subnet6Ptr.
-/// @tparam CollectionType One of the @c Subnet4Collection or @c Subnet6Collection.
+/// @tparam CollectionType One of the @c Subnet4SimpleCollection,
+/// @c Subnet4Collection, @c Subnet6SimpleCollection or @c Subnet6Collection.
 template<typename ReturnPtrType, typename CollectionType>
 class SubnetFetcher {
 public:
@@ -947,10 +1008,34 @@ using SubnetFetcher4 = SubnetFetcher<Subnet4Ptr, Subnet4Collection>;
 
 /// @brief Type of the @c SubnetFetcher used for IPv6.
 using SubnetFetcher6 = SubnetFetcher<Subnet6Ptr, Subnet6Collection>;
-
-
 //@}
 
+/// @brief adapters for linking templates to qualified names
+/// @{
+namespace {
+
+template <isc::util::DhcpSpace D>
+struct AdapterSubnet {};
+
+template <>
+struct AdapterSubnet<isc::util::DHCPv4> {
+    using type = Subnet4;
+};
+
+template <>
+struct AdapterSubnet<isc::util::DHCPv6> {
+    using type = Subnet6;
+};
+
+}  // namespace
+
+template <isc::util::DhcpSpace D>
+using SubnetT = typename AdapterSubnet<D>::type;
+
+template <isc::util::DhcpSpace D>
+using SubnetTPtr = boost::shared_ptr<SubnetT<D>>;
+/// @}
+
 } // end of isc::dhcp namespace
 } // end of isc namespace
 
