@@ -1,4 +1,4 @@
-// Copyright (C) 2009-2021 Internet Systems Consortium, Inc. ("ISC")
+// Copyright (C) 2009-2022 Internet Systems Consortium, Inc. ("ISC")
 //
 // This Source Code Form is subject to the terms of the Mozilla Public
 // License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -282,7 +282,7 @@ protected:
 ///
 /// Note to developers: it may make more sense to introduce an abstract base
 /// class for the \c OutputBuffer and define the simple implementation as a
-/// a concrete derived class.  That way we can provide flexibility for future
+/// concrete derived class.  That way we can provide flexibility for future
 /// extension such as more efficient buffer implementation or allowing users
 /// to have their own customized version without modifying the source code.
 /// We in fact considered that option, but at the moment chose the simpler
@@ -548,6 +548,10 @@ public:
     /// \param data A pointer to the data to be copied into the buffer.
     /// \param len The length of the data in bytes.
     void writeData(const void *data, size_t len) {
+        if (len == 0) {
+            return;
+        }
+
         ensureAllocated(size_ + len);
         static_cast<void>(std::memmove(buffer_ + size_, data, len));
         size_ += len;
