@@ -22,7 +22,7 @@ prefix="/usr/local"
 if test -f "${prefix}/share/kea/scripts/admin-utils.sh"; then
     . "${prefix}/share/kea/scripts/admin-utils.sh"
 else
-    . "/pobj/kea-2.4.1-mysql/kea-2.4.1/src/bin/admin/admin-utils.sh"
+    . "/local/pobj/kea-2.4.2-mysql/kea-2.4.2/src/bin/admin/admin-utils.sh"
 fi
 
 # Check version.
